"""
PC-BASIC - values.py
Types, values and conversions

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import math
import struct
import functools

from ...compat import int2byte

from ..base import error
from ..base import tokens as tk
from . import numbers
from . import strings


# BASIC type sigils:
# Integer (%) - stored as two's complement, little-endian
# Single (!) - stored as 4-byte Microsoft Binary Format
# Double (#) - stored as 8-byte Microsoft Binary Format
# String ($) - stored as 1-byte length plus 2-byte pointer to string space
INT = numbers.Integer.sigil
SNG = numbers.Single.sigil
DBL = numbers.Double.sigil
STR = strings.String.sigil

# storage size in bytes
TYPE_TO_SIZE = {STR: 3, INT: 2, SNG: 4, DBL: 8}
SIZE_TO_TYPE = {2: INT, 3: STR, 4: SNG, 8: DBL}

# type classes
SIZE_TO_CLASS = {
    2: numbers.Integer,
    3: strings.String,
    4: numbers.Single,
    8: numbers.Double
}

TYPE_TO_CLASS = {
    INT: numbers.Integer,
    STR: strings.String,
    SNG: numbers.Single,
    DBL: numbers.Double
}

# cutoff for trigonometric functions
# above this machine precision makes the result useless and machine/os dependent
# this is close to what gw uses but not quite equivalent
TRIG_MAX = 5e16

def size_bytes(name):
    """Return the size of a value type, by variable name or type char."""
    return TYPE_TO_SIZE[name[-1:]]

###############################################################################
# type checks

def check_value(inp):
    """Check if value is of Value type."""
    if not isinstance(inp, numbers.Value):
        raise TypeError('%s is not of class Value' % type(inp))

def pass_string(inp, err=error.TYPE_MISMATCH):
    """Check if variable is String-valued."""
    if not isinstance(inp, strings.String):
        check_value(inp)
        raise error.BASICError(err)
    return inp

def pass_number(inp, err=error.TYPE_MISMATCH):
    """Check if variable is numeric."""
    if not isinstance(inp, numbers.Number):
        check_value(inp)
        raise error.BASICError(err)
    return inp

def next_string(args):
    """Retrieve a string from an iterator and return as Python value."""
    expr = next(args)
    return to_string_or_none(expr)

def to_string_or_none(expr):
    if isinstance(expr, strings.String):
        return expr.to_value()
    elif expr is None:
        return expr
    else:
        raise error.BASICError(error.TYPE_MISMATCH)


###############################################################################
# type conversions

def match_types(left, right):
    """Check if variables are numeric and convert to highest-precision."""
    if isinstance(left, numbers.Double) or isinstance(right, numbers.Double):
        return to_double(left), to_double(right)
    elif isinstance(left, numbers.Single) or isinstance(right, numbers.Single):
        return to_single(left), to_single(right)
    elif isinstance(left, numbers.Integer) or isinstance(right, numbers.Integer):
        return to_integer(left), to_integer(right)
    elif isinstance(left, strings.String) or isinstance(right, strings.String):
        return pass_string(left), pass_string(right)
    raise TypeError('%s or %s is not of class Value.' % (type(left), type(right)))


###############################################################################
# error handling

def float_safe(fn):
    """Decorator to handle floating point errors."""
    def wrapped_fn(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except (ValueError, ArithmeticError) as e:
            return args[0].error_handler.handle(e)
    wrapped_fn.__name__ = fn.__name__
    return wrapped_fn

def _call_float_function(fn, *args):
    """Convert to IEEE 754, apply function, convert back."""
    args = list(args)
    pass_number(args[0])
    values = args[0]._values
    feh = args[0].error_handler
    try:
        # to_float can overflow on Double.pos_max
        args = [_arg.to_float(values.double_math) for _arg in args]
        floatcls = args[0].__class__
        python_args = [_arg.to_value() for _arg in args]
        result = fn(*python_args)
        # python3 may return complex values for some real functions
        # where python2 simply raises an error
        if isinstance(result, complex):
            raise ValueError('Non-real result')
        return floatcls(None, values).from_value(result)
    except (ValueError, ArithmeticError) as e:
        # create positive infinity of the appropriate class
        if values.double_math and isinstance(args[0], numbers.Double):
            floatcls = numbers.Double
        else:
            floatcls = numbers.Single
        infty = floatcls(None, values).from_bytes(floatcls.pos_max)
        # attach as exception payload for float error handler to deal with
        return feh.handle(e.__class__(infty))


class FloatErrorHandler(object):
    """Handles floating point errors."""

    # types of errors that do not always interrupt execution
    soft_types = (error.OVERFLOW, error.DIVISION_BY_ZERO)

    def __init__(self, console):
        """Setup handler."""
        self._console = console
        self._do_raise = False

    def suspend(self, do_raise):
        """Pause local handling of floating point errors."""
        self._do_raise = do_raise

    def handle(self, e):
        """Handle Overflow or Division by Zero."""
        if isinstance(e, ValueError):
            # math domain errors such as SQR(-1)
            math_error = error.IFC
        elif isinstance(e, OverflowError):
            math_error = error.OVERFLOW
        elif isinstance(e, ZeroDivisionError):
            math_error = error.DIVISION_BY_ZERO
        else: # pragma: no cover
            # shouldn't happen, we're only called with ValueError/ArithmeticError
            raise e
        if (self._do_raise or self._console is None or math_error not in self.soft_types):
            # also raises exception in error_handle_mode!
            # in that case, prints a normal error message
            raise error.BASICError(math_error)
        else:
            # write a message & continue as normal
            # message should not include line number or trailing \xFF
            self._console.write_line(error.BASICError(math_error).message)
        # return max value for the appropriate float type
        # integer operations should just raise the BASICError directly, they are not handled
        if e.args and isinstance(e.args[0], numbers.Float):
            return e.args[0]
        else: # pragma: no cover
            return numbers.Single(None, self).from_bytes(numbers.Single.pos_max)


###############################################################################

class Values(object):
    """Handles BASIC strings and numbers."""

    def __init__(self, string_space, double_math):
        """Setup values."""
        self.stringspace = string_space
        # double-precision EXP, SIN, COS, TAN, ATN, LOG
        self.double_math = double_math
        self.error_handler = None

    def set_handler(self, handler):
        """Initialise the error message console."""
        self.error_handler = handler

    def create(self, buf):
        """Create new variable object with buffer provided."""
        # this sets a view, not a copy
        return SIZE_TO_CLASS[len(buf)](buf, self)

    def new(self, sigil):
        """Return newly allocated value of the given type with zeroed buffer."""
        return TYPE_TO_CLASS[sigil](None, self)

    def new_string(self):
        """Return newly allocated null string."""
        return strings.String(None, self)

    def new_integer(self):
        """Return newly allocated zero integer."""
        return numbers.Integer(None, self)

    def new_single(self):
        """Return newly allocated zero single."""
        return numbers.Single(None, self)

    def new_double(self):
        """Return newly allocated zero double."""
        return numbers.Double(None, self)

    ###########################################################################
    # convert between BASIC and Python values

    @float_safe
    def from_value(self, python_val, typechar):
        """Convert Python value to BASIC value."""
        return TYPE_TO_CLASS[typechar](None, self).from_value(python_val)

    def from_str_at(self, python_str, address):
        """Convert str to String at given address."""
        return strings.String(None, self).from_pointer(
            *self.stringspace.store(python_str, address))

    def from_bool(self, boo):
        """Convert Python boolean to Integer."""
        if boo:
            return numbers.Integer(None, self).from_bytes(b'\xff\xff')
        return numbers.Integer(None, self)

    ###########################################################################
    # convert to and from internal representation

    def from_bytes(self, token_bytes):
        """Convert internal byte representation to BASIC value."""
        # make a copy, not a view
        return SIZE_TO_CLASS[len(token_bytes)](None, self).from_bytes(token_bytes)

    def from_token(self, token):
        """Convert number token to new Number temporary"""
        if not token:
            raise ValueError('Token must not be empty')
        lead = bytes(token)[0:1]
        if lead == tk.T_SINGLE:
            return numbers.Single(None, self).from_token(token)
        elif lead == tk.T_DOUBLE:
            return numbers.Double(None, self).from_token(token)
        elif lead in tk.NUMBER:
            return numbers.Integer(None, self).from_token(token)
        raise ValueError('%s is not a number token' % repr(token))

    ###########################################################################
    # create value from string representations

    @float_safe
    def from_repr(self, word, allow_nonnum, typechar=None):
        """Convert representation to value."""
        # keep as string if typechar asks for it, ignore typechar otherwise
        if typechar == STR:
            return self.new_string().from_str(word)
        # skip spaces and line feeds (but not NUL).
        word = word.lstrip(b' \n').upper()
        if not word:
            return self.new_integer()
        if word[:2] == b'&H':
            return self.new_integer().from_hex(word[2:])
        elif word[:1] == b'&':
            return self.new_integer().from_oct(word[2:] if word[1:2] == b'O' else word[1:])
        # we need to try to convert to int first,
        # mainly so that the tokeniser can output the right token type
        try:
            return self.new_integer().from_str(word)
        except ValueError as e:
            # non-integer characters, try a float
            pass
        except error.BASICError as e:
            if e.err != error.OVERFLOW: # pragma: no cover
                # shouldn't happen, from_str only raises Overflow
                raise
        # if allow_nonnum == False, raises ValueError for non-numerical characters
        is_double, mantissa, exp10 = numbers.str_to_decimal(word, allow_nonnum)
        if is_double:
            return self.new_double().from_decimal(mantissa, exp10)
        return self.new_single().from_decimal(mantissa, exp10)


###############################################################################
# conversions

def to_integer(inp, unsigned=False):
    """Check if variable is numeric, convert to Int."""
    if isinstance(inp, strings.String):
        raise error.BASICError(error.TYPE_MISMATCH)
    return inp.to_integer(unsigned)

@float_safe
def to_single(num):
    """Check if variable is numeric, convert to Single."""
    if isinstance(num, strings.String):
        raise error.BASICError(error.TYPE_MISMATCH)
    return num.to_single()

@float_safe
def to_double(num):
    """Check if variable is numeric, convert to Double."""
    if isinstance(num, strings.String):
        raise error.BASICError(error.TYPE_MISMATCH)
    return num.to_double()

def cint_(args):
    """CINT: convert to integer (by rounding, halves away from zero)."""
    value, = args
    return to_integer(value)

def csng_(args):
    """CSNG: convert to single (by Gaussian rounding)."""
    value, = args
    return to_single(value)

def cdbl_(args):
    """CDBL: convert to double."""
    value, = args
    return to_double(value)

def to_type(typechar, value):
    """Check if variable can be converted to the given type and convert if necessary."""
    if typechar == STR:
        return pass_string(value)
    elif typechar == INT:
        return to_integer(value)
    elif typechar == SNG:
        return to_single(value)
    elif typechar == DBL:
        return to_double(value)
    raise ValueError('%s is not a valid sigil.' % typechar)

# NOTE that this function will overflow if outside the range of Integer
# whereas Float.to_int will not
def to_int(inp, unsigned=False):
    """Round numeric variable and convert to Python integer."""
    return to_integer(inp, unsigned).to_int(unsigned)

def mki_(args):
    """MKI$: return the byte representation of an int."""
    x, = args
    return x._values.new_string().from_str(bytes(to_integer(x).to_bytes()))

def mks_(args):
    """MKS$: return the byte representation of a single."""
    x, = args
    return x._values.new_string().from_str(bytes(to_single(x).to_bytes()))

def mkd_(args):
    """MKD$: return the byte representation of a double."""
    x, = args
    return x._values.new_string().from_str(bytes(to_double(x).to_bytes()))

def cvi_(args):
    """CVI: return the int value of a byte representation."""
    x, = args
    cstr = pass_string(x).to_str()
    error.throw_if(len(cstr) < 2)
    return x._values.from_bytes(cstr[:2])

def cvs_(args):
    """CVS: return the single-precision value of a byte representation."""
    x, = args
    cstr = pass_string(x).to_str()
    error.throw_if(len(cstr) < 4)
    return x._values.from_bytes(cstr[:4])

def cvd_(args):
    """CVD: return the double-precision value of a byte representation."""
    x, = args
    cstr = pass_string(x).to_str()
    error.throw_if(len(cstr) < 8)
    return x._values.from_bytes(cstr[:8])


###############################################################################
# comparisons

def _bool_eq(left, right):
    """Return true if left == right, false otherwise."""
    left, right = match_types(left, right)
    return left.eq(right)

def _bool_gt(left, right):
    """Ordering: return -1 if left > right, 0 otherwise."""
    left, right = match_types(left, right)
    return left.gt(right)

def eq(left, right):
    """Return -1 if left == right, 0 otherwise."""
    return left._values.from_bool(_bool_eq(left, right))

def neq(left, right):
    """Return -1 if left != right, 0 otherwise."""
    return left._values.from_bool(not _bool_eq(left, right))

def gt(left, right):
    """Ordering: return -1 if left > right, 0 otherwise."""
    return left._values.from_bool(_bool_gt(left, right))

def gte(left, right):
    """Ordering: return -1 if left >= right, 0 otherwise."""
    return left._values.from_bool(not _bool_gt(right, left))

def lte(left, right):
    """Ordering: return -1 if left <= right, 0 otherwise."""
    return left._values.from_bool(not _bool_gt(left, right))

def lt(left, right):
    """Ordering: return -1 if left < right, 0 otherwise."""
    return left._values.from_bool(_bool_gt(right, left))


###############################################################################
# bitwise operators

def not_(num):
    """Bitwise NOT, -x-1."""
    return num._values.new_integer().from_int(~to_integer(num).to_int())

def and_(left, right):
    """Bitwise AND."""
    return left._values.new_integer().from_int(
        to_integer(left).to_int(unsigned=True) & to_integer(right).to_int(unsigned=True),
        unsigned=True)

def or_(left, right):
    """Bitwise OR."""
    return left._values.new_integer().from_int(
        to_integer(left).to_int(unsigned=True) | to_integer(right).to_int(unsigned=True),
        unsigned=True)

def xor_(left, right):
    """Bitwise XOR."""
    return left._values.new_integer().from_int(
        to_integer(left).to_int(unsigned=True) ^ to_integer(right).to_int(unsigned=True),
        unsigned=True)

def eqv_(left, right):
    """Bitwise equivalence."""
    return left._values.new_integer().from_int(
        ~(to_integer(left).to_int(unsigned=True) ^ to_integer(right).to_int(unsigned=True)),
        unsigned=True)

def imp_(left, right):
    """Bitwise implication."""
    return left._values.new_integer().from_int(
        (~to_integer(left).to_int(unsigned=True)) | right.to_integer().to_int(unsigned=True),
        unsigned=True)


##############################################################################
# unary operations

def abs_(args):
    """Return the absolute value of a number. No-op for strings."""
    inp, = args
    if isinstance(inp, strings.String):
        # strings pass unchanged
        return inp
    # promote Integer to Single to avoid integer overflow on -32768
    return inp.to_float().clone().iabs()

def neg(inp):
    """Negation (unary -). No-op for strings."""
    if isinstance(inp, strings.String):
        # strings pass unchanged
        return inp
    # promote Integer to Single to avoid integer overflow on -32768
    return inp.to_float().clone().ineg()

def sgn_(args):
    """Sign."""
    x, = args
    return numbers.Integer(None, x._values).from_int(pass_number(x).sign())

def int_(args):
    """Truncate towards negative infinity (INT)."""
    inp, = args
    if isinstance(inp, strings.String):
        # strings pass unchanged
        return inp
    return inp.clone().ifloor()

def fix_(args):
    """Truncate towards zero."""
    inp, = args
    return pass_number(inp).clone().itrunc()

def sqr_(args):
    """Square root."""
    x, = args
    return _call_float_function(math.sqrt, x)

def exp_(args):
    """Exponential."""
    x, = args
    return _call_float_function(math.exp, x)

def sin_(args):
    """Sine."""
    x, = args
    return _call_float_function(lambda _x: math.sin(_x) if abs(_x) < TRIG_MAX else 0., x)

def cos_(args):
    """Cosine."""
    x, = args
    return _call_float_function(lambda _x: math.cos(_x) if abs(_x) < TRIG_MAX else 1., x)

def tan_(args):
    """Tangent."""
    x, = args
    return _call_float_function(lambda _x: math.tan(_x) if abs(_x) < TRIG_MAX else 0., x)

def atn_(args):
    """Inverse tangent."""
    x, = args
    return _call_float_function(math.atan, x)

def log_(args):
    """Logarithm."""
    x, = args
    return _call_float_function(math.log, x)


######################################################################
# string representations and characteristics

def to_repr(inp, leading_space, type_sign):
    """Convert BASIC number to Python bytes representation."""
    # PRINT, STR$ - yes leading space, no type sign
    # WRITE - no leading space, no type sign
    # LIST - no loading space, yes type sign
    if isinstance(inp, numbers.Number):
        return inp.to_str(leading_space, type_sign)
    elif isinstance(inp, strings.String):
        raise error.BASICError(error.TYPE_MISMATCH)
    raise TypeError('%s is not of class Value' % type(inp))

def str_(args):
    """STR$: string representation of a number."""
    x, = args
    return x._values.new_string().from_str(
        to_repr(pass_number(x), leading_space=True, type_sign=False)
    )

def val_(args):
    """VAL: number value of a string."""
    x, = args
    return x._values.from_repr(pass_string(x).to_str(), allow_nonnum=True)

def len_(args):
    """LEN: length of string."""
    s, = args
    return pass_string(s).len()

def space_(args):
    """SPACE$: repeat spaces."""
    num, = args
    return num._values.new_string().space(pass_number(num))

def asc_(args):
    """ASC: ordinal ASCII value of a character."""
    s, = args
    return pass_string(s).asc()

def chr_(args):
    """CHR$: character for ASCII value."""
    x, = args
    val = pass_number(x).to_integer().to_int()
    error.range_check(0, 255, val)
    return x._values.new_string().from_str(int2byte(val))

def oct_(args):
    """OCT$: octal representation of int."""
    x, = args
    # allow range -32768 to 65535
    val = to_integer(x, unsigned=True)
    return x._values.new_string().from_str(val.to_oct())

def hex_(args):
    """HEX$: hexadecimal representation of int."""
    x, = args
    # allow range -32768 to 65535
    val = to_integer(x, unsigned=True)
    return x._values.new_string().from_str(val.to_hex())

##############################################################################
# sring operations

def left_(args):
    """LEFT$: get substring of num characters at the start of string."""
    s, num = next(args), next(args)
    s, num = pass_string(s), to_integer(num)
    list(args)
    stop = num.to_integer().to_int()
    if stop == 0:
        return s.new()
    error.range_check(0, 255, stop)
    return s.new().from_str(s.to_str()[:stop])

def right_(args):
    """RIGHT$: get substring of num characters at the end of string."""
    s, num = next(args), next(args)
    s, num = pass_string(s), to_integer(num)
    list(args)
    stop = num.to_integer().to_int()
    if stop == 0:
        return s.new()
    error.range_check(0, 255, stop)
    return s.new().from_str(s.to_str()[-stop:])

def mid_(args):
    """MID$: get substring."""
    s, start = next(args), to_integer(next(args))
    p = pass_string(s)
    num = next(args)
    if num is not None:
        num = to_integer(num)
    list(args)
    length = s.length()
    start = start.to_integer().to_int()
    if num is None:
        num = length
    else:
        num = num.to_integer().to_int()
    error.range_check(1, 255, start)
    error.range_check(0, 255, num)
    if num == 0 or start > length:
        return s.new()
    # BASIC's indexing starts at 1, Python's at 0
    start -= 1
    return s.new().from_str(s.to_str()[start:start+num])

def instr_(args):
    """INSTR: find substring in string."""
    arg0 = next(args)
    if isinstance(arg0, numbers.Number):
        start = to_int(arg0)
        error.range_check(1, 255, start)
        big = pass_string(next(args))
    else:
        start = 1
        big = pass_string(arg0)
    small = pass_string(next(args))
    list(args)
    new_int = numbers.Integer(None, big._values)
    big = big.to_str()
    small = small.to_str()
    if big == b'' or start > len(big):
        return new_int
    # BASIC counts string positions from 1
    find = big[start-1:].find(small)
    if find == -1:
        return new_int
    return new_int.from_int(start + find)

def string_(args):
    """STRING$: repeat a character num times."""
    num = to_int(next(args))
    error.range_check(0, 255, num)
    asc_value_or_char = next(args)
    if isinstance(asc_value_or_char, numbers.Integer):
        error.range_check(0, 255, asc_value_or_char.to_int())
    list(args)
    if isinstance(asc_value_or_char, strings.String):
        char = asc_value_or_char.to_str()[:1]
    else:
        # overflow if outside Integer range
        ascval = asc_value_or_char.to_integer().to_int()
        error.range_check(0, 255, ascval)
        char = int2byte(ascval)
    return strings.String(None, asc_value_or_char._values).from_str(char * num)

##############################################################################
# binary operations

@float_safe
def pow(left, right):
    """Left^right."""
    if isinstance(left, strings.String) or isinstance(right, strings.String):
        raise error.BASICError(error.TYPE_MISMATCH)
    if left._values.double_math and (
            isinstance(left, numbers.Double) or isinstance(right, numbers.Double)
        ):
        return _call_float_function(lambda a, b: a**b, to_double(left), to_double(right))
    elif isinstance(right, numbers.Integer):
        return left.to_single().clone().ipow_int(right)
    else:
        return _call_float_function(lambda a, b: a**b, to_single(left), to_single(right))

@float_safe
def add(left, right):
    """Add two numbers or concatenate two strings."""
    if isinstance(left, numbers.Number):
        # promote Integer to Single to avoid integer overflow
        left = left.to_float()
    left, right = match_types(left, right)
    # note that we can't call iadd here, as it breaks with strings
    # since between copy and dereference the address may change due to garbage collection
    # it may be better to define non-in-place operators for everything
    return left.add(right)

@float_safe
def sub(left, right):
    """Subtract two numbers."""
    if isinstance(left, strings.String) or isinstance(right, strings.String):
        raise error.BASICError(error.TYPE_MISMATCH)
    # promote Integer to Single to avoid integer overflow
    left, right = match_types(left.to_float(), right)
    return left.clone().isub(right)


@float_safe
def mul(left, right):
    """Left*right."""
    if isinstance(left, strings.String) or isinstance(right, strings.String):
        raise error.BASICError(error.TYPE_MISMATCH)
    elif isinstance(left, numbers.Double) or isinstance(right, numbers.Double):
        return left.to_double().clone().imul(right.to_double())
    else:
        return left.to_single().clone().imul(right.to_single())

@float_safe
def div(left, right):
    """Left/right."""
    if isinstance(left, strings.String) or isinstance(right, strings.String):
        raise error.BASICError(error.TYPE_MISMATCH)
    elif isinstance(left, numbers.Double) or isinstance(right, numbers.Double):
        return left.to_double().clone().idiv(right.to_double())
    else:
        return left.to_single().clone().idiv(right.to_single())

@float_safe
def intdiv(left, right):
    """Left\\right."""
    return to_integer(left).clone().idiv_int(to_integer(right))

@float_safe
def mod_(left, right):
    """Left modulo right."""
    return to_integer(left).clone().imod(to_integer(right))


# conversions to type
TYPE_TO_CONV = {STR: pass_string, INT: to_integer, SNG: to_single, DBL: to_double}



"""
PC-BASIC - dos.py
Operating system shell and environment

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os
import io
import sys
import logging
import threading
import time
from collections import deque
import subprocess
from subprocess import Popen, PIPE

from ..compat import OEM_ENCODING, HIDE_WINDOW, PY2
from ..compat import which, split_quoted, getenvu, setenvu, iterenvu
from .codepage import CONTROL
from .base import error
from . import values


# command interpreter must support command.com convention
# to be able to use SHELL "dos-command"
# on linux, "wine cmd.exe" works OK
# dosemu can probably be made to work with a wrapper script
# sh doesn't work but makes little sense to use anyway as it's totally unlike MS-DOS
SHELL_COMMAND_SWITCH = u'/C'


##########################################
# calling shell environment

class Environment(object):
    """Handle environment changes."""

    def __init__(self, values, codepage):
        """Initialise."""
        self._values = values
        self._codepage = codepage

    def _setenv(self, key, value):
        """Set environment (bytes) key to (bytes) value."""
        assert isinstance(key, bytes), type(key)
        assert isinstance(value, bytes), type(value)
        # only accept ascii-128 for keys
        try:
            ukey = key.decode('ascii')
        except UnicodeError:
            raise error.BASICError(error.IFC)
        # enforce uppercase
        ukey = ukey.upper()
        uvalue = self._codepage.bytes_to_unicode(value)
        setenvu(ukey, uvalue)

    def _getenv(self, key):
        """Get environment (bytes) value or b''."""
        assert isinstance(key, bytes), type(key)
        # only accept ascii-128 for keys
        try:
            ukey = key.decode('ascii')
        except UnicodeError:
            raise error.BASICError(error.IFC)
        # enforce uppercase
        ukey = ukey.upper()
        return self._codepage.unicode_to_bytes(getenvu(ukey, u''))

    def _getenv_item(self, index):
        """Get environment (bytes) 'key=value' or b'', by zero-based index."""
        envlist = list(iterenvu())
        try:
            ukey = envlist[index]
        except IndexError:
            return b''
        else:
            key = self._codepage.unicode_to_bytes(ukey)
            value = self._codepage.unicode_to_bytes(getenvu(ukey, u''))
            return b'%s=%s' % (key, value)

    def environ_(self, args):
        """ENVIRON$: get environment string."""
        expr, = args
        if isinstance(expr, values.String):
            key = expr.to_str()
            if not key:
                raise error.BASICError(error.IFC)
            result = self._getenv(key)
        else:
            index = values.to_int(expr)
            error.range_check(1, 255, index)
            result = self._getenv_item(index-1)
        return self._values.new_string().from_str(result)

    def environ_statement_(self, args):
        """ENVIRON: set environment string."""
        envstr = values.next_string(args)
        list(args)
        eqs = envstr.find(b'=')
        if eqs <= 0:
            raise error.BASICError(error.IFC)
        self._setenv(envstr[:eqs], envstr[eqs+1:])


#########################################
# shell

class Shell(object):
    """Launcher for command shell."""

    def __init__(self, queues, keyboard, console, files, codepage, shell):
        """Initialise the shell."""
        self._shell = shell
        self._queues = queues
        self._keyboard = keyboard
        self._console = console
        self._files = files
        self._codepage = codepage
        self._last_command = u''

    if PY2: # pragma: no cover
        def _process_stdout(self, stream, output):
            """Retrieve SHELL output and write to console."""
            # hack for python 2: use latin-1 as a passthrough encoding
            # doesn't work correctly but so be it
            stream = io.open(stream.fileno(), mode='r', encoding='latin-1', closefd=False)
            while True:
                # blocking read
                c = stream.read(1)
                # hack for python 2 - ignore NULs so utf-16 encoded ascii comes through sort of ok
                if c == b'\0':
                    continue
                # stream ends if process closes
                if not c:
                    return
                # don't access console in this thread
                # the other thread already does
                output.append(c)
            stream.close()
    else:
        def _process_stdout(self, stream, output):
            """Retrieve SHELL output and write to console."""
            first = stream.peek(2)
            # detect utf-16 (windows unicode shell)
            encoding = 'utf-16le' if first[1:2] == b'\0' else OEM_ENCODING
            stream = io.TextIOWrapper(stream, encoding=encoding, errors='replace')
            while True:
                # blocking read
                c = stream.read(1)
                # stream ends if process closes
                if not c:
                    return
                # don't access console in this thread
                # the other thread already does
                output.append(c)

    def launch(self, command):
        """Run a SHELL subprocess."""
        logging.debug('Executing SHELL command `%r` with command interpreter `%s`', command, self._shell)
        if not self._shell:
            logging.warning('SHELL statement not enabled: no command interpreter specified.')
            raise error.BASICError(error.IFC)
        # split by whitespace
        cmd = split_quoted(self._shell, quote=u'"', strip_quotes=True)
        # find executable on path
        cmd[0] = which(cmd[0], path='.' + os.pathsep + os.environ.get("PATH"))
        if not cmd[0]:
            logging.warning(u'SHELL: command interpreter `%s` not found.', self._shell)
            raise error.BASICError(error.IFC)
        if command:
            cmd += [SHELL_COMMAND_SWITCH, self._codepage.bytes_to_unicode(command, box_protect=False)]
        # get working directory; also raises IFC if current_device is CAS1
        work_dir = self._files.get_native_cwd() or '.'
        try:
            p = Popen(
                cmd, shell=False, cwd=work_dir,
                stdin=PIPE, stdout=PIPE, stderr=PIPE,
                # first try with HIDE_WINDOW to avoid ugly command window popping up on windows
                startupinfo=HIDE_WINDOW
            )
        except EnvironmentError:
            try:
                # HIDE_WINDOW not allowed on Windows when called from console
                p = Popen(
                    cmd, shell=False, cwd=work_dir,
                    stdin=PIPE, stdout=PIPE, stderr=PIPE,
                )
            except EnvironmentError as err:
                logging.warning(
                    u'SHELL: command interpreter `%s` not accessible: %s', self._shell, err
                )
                raise error.BASICError(error.IFC)
        shell_output = self._launch_reader_thread(p.stdout)
        shell_cerr = self._launch_reader_thread(p.stderr)
        try:
            self._communicate(p, shell_output, shell_cerr)
        except EnvironmentError as e:
            logging.warning(e)
        finally:
            # ensure the process is terminated on exit
            if p.poll() is None:
                p.kill()

    def _launch_reader_thread(self, stream):
        """Launch output reader."""
        shell_output = deque()
        outp = threading.Thread(target=self._process_stdout, args=(stream, shell_output))
        # daemonise or join later? if we join, a shell that doesn't close will hang us on exit
        outp.daemon = True
        outp.start()
        return shell_output

    def _drain_final(self, shell_output, remove_echo):
        """Drain final output from shell."""
        if not shell_output:
            return
        if not shell_output[-1] == u'\n':
            shell_output.append(u'\n')
        self._show_output(shell_output, remove_echo)

    def _communicate(self, p, shell_output, shell_cerr):
        """Communicate with launched shell."""
        word = []
        while p.poll() is None:
            # stderr output should come first
            # e.g. first print the error message (tsderr), then the prompt (stdout)
            self._show_output(shell_cerr, remove_echo=False)
            self._show_output(shell_output, remove_echo=True)
            try:
                self._queues.wait()
                # expand=False suppresses key macros
                c = self._keyboard.get_fullchar(expand=False)
            except error.Break:
                pass
            if not c:
                continue
            elif c in (b'\r', b'\n'):
                self._console.write(c)
                # send the command
                self._send_input(p.stdin, word)
                word = []
            elif c == b'\b':
                # handle backspace
                if word:
                    word.pop()
                    self._console.write(b'\x1D \x1D')
            elif not c.startswith(b'\0'):
                # exclude e-ascii (arrow keys not implemented)
                word.append(c)
                self._console.write(c)
        # drain final output
        self._drain_final(shell_cerr, remove_echo=False)
        self._drain_final(shell_output, remove_echo=True)

    def _send_input(self, pipe, word):
        """Write keyboard input to pipe."""
        word.extend([b'\r', b'\n'])
        bytes_word = b''.join(word)
        unicode_word = self._codepage.bytes_to_unicode(
            bytes_word, preserve=CONTROL, box_protect=False
        )
        # match universal newlines
        self._last_command = unicode_word.replace(u'\r\n', u'\n').replace(u'\r', u'\n')
        logging.debug('SHELL << %r', unicode_word)
        # cmd.exe /u outputs UTF-16 but does not accept it as input...
        shell_word = unicode_word.encode(OEM_ENCODING, errors='replace')
        pipe.write(shell_word)
        # explicit flush as pipe may not be line buffered. blocks in python 3 without
        pipe.flush()

    def _show_output(self, shell_output, remove_echo):
        """Write shell output to console."""
        if shell_output and u'\n' in shell_output:
            # can't do a comprehension as it will crash if the deque is accessed by the thread
            chars = deque()
            while shell_output:
                chars.append(shell_output.popleft())
            outstr = u''.join(chars)
            logging.debug('SHELL >> %r', outstr)
            # detect echo
            if remove_echo:
                outstr = self._remove_echo(outstr)
            # remove BELs (dosemu uses these a lot)
            outstr = outstr.replace(u'\x07', u'')
            # use BASIC newlines (CR) instead of universal newlines (LF)
            outstr = outstr.replace(u'\n', u'\r')
            # encode to codepage
            outbytes = self._codepage.unicode_to_bytes(outstr, errors='replace')
            self._console.write(outbytes)

    def _remove_echo(self, unicode_reply):
        """Detect if output was an echo of the input and remove."""
        if unicode_reply.startswith(self._last_command):
            unicode_reply = unicode_reply[len(self._last_command):]
        self._last_command = u''
        return unicode_reply



"""
PC-BASIC - interface.base
Interface utility classes

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os


# message displayed when wiating to close
WAIT_MESSAGE = u'Press a key to close window'
# message displayed when Alt-F4 inhibited
NOKILL_MESSAGE = u'to exit type <CTRL+BREAK> <ESC> SYSTEM'


class InitFailed(Exception):
    """Initialisation failed."""


class PluginRegister(object):
    """Plugin register."""

    def __init__(self):
        """Initialise plugin register."""
        self._plugins = {}

    def register(self, name):
        """Decorator to register a plugin."""
        def decorated_plugin(cls):
            self._plugins[name] = cls
            return cls
        return decorated_plugin

    def __getitem__(self, name):
        """Retrieve plugin."""
        return self._plugins[name]


class EnvironmentCache(object):
    """ Set environment variables for temporary use and clean up nicely."""

    def __init__(self):
        """Create the environment cache."""
        self._saved = {}

    def __enter__(self):
        """Context guard."""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Context guard."""
        self.close()

    def set(self, key, value):
        """Set an environment variable and save the original value in the cache."""
        if key in self._saved:
            self.reset(key)
        self._saved[key] = os.environ.get(key)
        os.environ[key] = value

    def reset(self, key):
        """Restore the original value for an environment variable."""
        cached = self._saved.pop(key, None)
        if cached is None:
            os.environ.pop(key, None)
        else:
            os.environ[key] = cached

    def close(self):
        """Restore all environment variables."""
        for key in list(self._saved.keys()):
            self.reset(key)

    def __del__(self):
        """Clean up the cache."""
        self.close()


###############################################################################
# plugin registers

video_plugins = PluginRegister()
audio_plugins = PluginRegister()



"""
PC-BASIC - GW-BASIC/BASICA/Cartridge BASIC compatible interpreter

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from .data import NAME, VERSION, LONG_VERSION, AUTHOR, COPYRIGHT
from .api import Session, codepage, font
from .base.error import *
from .base import signals, scancode, eascii

__version__ = VERSION



"""
PC-BASIC - program.py
Program buffer utilities

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import binascii
import logging
import struct
import io

from ..compat import int2byte

from .base import error
from .base import tokens as tk
from . import values
from . import converter


class Program(object):
    """BASIC program."""

    def __init__(self, tokeniser, lister, hide_listing,
                allow_protect, allow_code_poke, memory, bytecode, rebuild_offsets):
        """Initialise program."""
        self._memory = memory
        # program bytecode buffer
        self.bytecode = bytecode
        self.erase()
        self.max_list_line = hide_listing if hide_listing else 65535
        self.allow_protect = allow_protect
        self.allow_code_poke = allow_code_poke
        self._rebuild_offsets = rebuild_offsets
        # to be set when file memory is initialised
        self.code_start = memory.code_start
        # for detokenise_line()
        self.tokeniser = tokeniser
        self.lister = lister

    def __repr__(self):
        """Return a marked-up hex dump of the program (for debugging)."""
        code = self.bytecode.getvalue()
        offset_val, p = 0, 0
        output = []
        for key in sorted(self.line_numbers.keys())[1:]:
            next_p = self.line_numbers[key]
            offset, linum = code[p+1:p+3], code[p+3:p+5]
            last_offset = offset_val
            offset_val = struct.unpack('<H', offset)[0] - (self.code_start + 1)
            linum_val, = struct.unpack('<H', linum)
            output.append(b'%s %s (+%03d) %s [%05d] %s' % (
                binascii.hexlify(code[p:p+1]),
                binascii.hexlify(offset),
                offset_val - last_offset,
                binascii.hexlify(code[p+3:p+5]),
                linum_val,
                binascii.hexlify(code[p+5:next_p])
            ))
            p = next_p
        output.append(b'%s %s (ENDS) %s %s' % (
            binascii.hexlify(code[p:p+1]),
            binascii.hexlify(code[p+1:p+3]),
            binascii.hexlify(code[p+3:p+5]),
            binascii.hexlify(code[p+5:])
        ))
        return b'\n'.join(output).decode('ascii', 'replace')

    def size(self):
        """Size of code space """
        return self.code_size

    def erase(self):
        """Erase the program from memory."""
        self.bytecode.seek(0)
        self.bytecode.write(b'\0\0\0')
        self.protected = False
        self.line_numbers = {65536: 0}
        self.last_stored = None
        self.code_size = self.bytecode.tell()
        self.bytecode.truncate()

    def truncate(self, rest=b''):
        """Write bytecode and cut the program of beyond the current position."""
        self.bytecode.write(rest if rest else b'\0\0\0')
        self.bytecode.truncate()
        # cut off at current position
        self.code_size = self.bytecode.tell()

    def explicit_lines(self, *line_range):
        """Convert iterables of lines with '.' into explicit numbers."""
        return (self.last_stored if l == b'.' else l for l in line_range)

    def get_line_number(self, pos):
        """Get line number for stream position."""
        pre = -1
        if pos is None:
            pos = -1
        for linum in self.line_numbers:
            linum_pos = self.line_numbers[linum]
            if linum_pos <= pos and linum > pre:
                pre = linum
        return pre

    def rebuild_line_dict(self):
        """Preparse to build line number dictionary."""
        self.line_numbers, offsets = {}, []
        self.bytecode.seek(0)
        scanline, scanpos, last = 0, 0, 0
        while True:
            # pass \x00
            self.bytecode.read(1)
            scanline = self.lister.detokenise_line_number(self.bytecode)
            if scanline == -1:
                scanline = 65536
                # if detokenise_line_number returns -1, it leaves the stream pointer here:
                # 00 _00_ 00 1A
                break
            self.line_numbers[scanline] = scanpos
            last = scanpos
            self.bytecode.skip_to(tk.END_LINE)
            scanpos = self.bytecode.tell()
            offsets.append(scanpos)
        self.line_numbers[65536] = scanpos
        # rebuild offsets
        if self._rebuild_offsets:
            self.bytecode.seek(0)
            last = 0
            for pos in offsets:
                self.bytecode.read(1)
                self.bytecode.write(struct.pack('<H', self.code_start + 1 + pos))
                self.bytecode.read(pos - last - 3)
                last = pos
            # ensure program is properly sealed - last offset must be 00 00.
            # keep, but ignore, anything after.
            self.bytecode.write(b'\0\0\0')

    def update_line_dict(self, pos, afterpos, length, deleteable, beyond):
        """Update line number dictionary after deleting lines."""
        # subtract length of line we replaced
        length -= afterpos - pos
        addr = (self.code_start + 1) + afterpos
        self.bytecode.seek(afterpos + length + 1)  # pass \x00
        while True:
            next_addr = self.bytecode.read(2)
            if len(next_addr) < 2 or next_addr == b'\0\0':
                break
            next_addr, = struct.unpack('<H', next_addr)
            self.bytecode.seek(-2, 1)
            self.bytecode.write(struct.pack('<H', next_addr + length))
            self.bytecode.read(next_addr - addr - 2)
            addr = next_addr
        # update line number dict
        for key in deleteable:
            del self.line_numbers[key]
        for key in beyond:
            self.line_numbers[key] += length

    def check_number_start(self, linebuf):
        """Check if the given line buffer starts with a line number."""
        # get the new line number
        linebuf.seek(1)
        scanline = self.lister.detokenise_line_number(linebuf)
        c = linebuf.skip_blank_read()
        # check if linebuf is an empty line after the line number
        empty = (c in tk.END_LINE)
        # check if we start with a number
        if c in tk.NUMBER:
            raise error.BASICError(error.STX)
        return empty, scanline

    def store_line(self, linebuf):
        """Store the given line buffer."""
        if self.protected:
            raise error.BASICError(error.IFC)
        # get the new line number
        linebuf.seek(1)
        scanline = self.lister.detokenise_line_number(linebuf)
        # check if linebuf is an empty line after the line number
        empty = (linebuf.skip_blank_read() in tk.END_LINE)
        pos, afterpos, deleteable, beyond = self.find_pos_line_dict(scanline, scanline)
        if empty and not deleteable:
            raise error.BASICError(error.UNDEFINED_LINE_NUMBER)
        # read the remainder of the program into a buffer to be pasted back after the write
        self.bytecode.seek(afterpos)
        rest = self.bytecode.read()
        # insert
        self.bytecode.seek(pos)
        # write the line buffer to the program buffer
        length = 0
        if not empty:
            # set offsets
            linebuf.seek(3) # pass \x00\xC0\xDE
            length = len(linebuf.getvalue())
            # check for free memory
            # variables are cleared upon program code storage
            if self.code_start + 1 + pos + length > self._memory.stack_start():
                raise error.BASICError(error.OUT_OF_MEMORY)
            self.bytecode.write(
                struct.pack('<BH', 0, self.code_start + 1 + pos + length) + linebuf.read()
            )
        # write back the remainder of the program
        self.truncate(rest)
        # update all next offsets by shifting them by the length of the added line
        self.update_line_dict(pos, afterpos, length, deleteable, beyond)
        if not empty:
            self.line_numbers[scanline] = pos
        self.last_stored = scanline

    def find_pos_line_dict(self, fromline, toline):
        """Find code positions for line range."""
        deleteable = [ num for num in self.line_numbers if num >= fromline and num <= toline ]
        beyond = [num for num in self.line_numbers if num > toline ]
        # find lowest number strictly above range
        afterpos = self.line_numbers[min(beyond)]
        # find lowest number within range
        try:
            startpos = self.line_numbers[min(deleteable)]
        except ValueError:
            startpos = afterpos
        return startpos, afterpos, deleteable, beyond

    def delete(self, fromline, toline):
        """Delete range of lines from stored program."""
        fromline, toline = self.explicit_lines(fromline, toline)
        fromline = fromline if fromline is not None else min(self.line_numbers)
        toline = toline if toline is not None else 65535
        startpos, afterpos, deleteable, beyond = self.find_pos_line_dict(fromline, toline)
        if not deleteable:
            # no lines selected
            raise error.BASICError(error.IFC)
        # do the delete
        self.bytecode.seek(afterpos)
        rest = self.bytecode.read()
        self.bytecode.seek(startpos)
        self.truncate(rest)
        # update line number dict
        self.update_line_dict(startpos, afterpos, 0, deleteable, beyond)

    def edit(self, console, from_line, target_bytepos):
        """Output program line to console and position cursor."""
        if self.protected:
            console.write(b'%d\r' % (from_line,))
            raise error.BASICError(error.IFC)
        # list line
        self.bytecode.seek(self.line_numbers[from_line]+1)
        _, output, byte_to_text_positions = self.lister.detokenise_line(self.bytecode)
        # find text position from byte position
        if target_bytepos is None:
            textpos = None
        else:
            textpos = min(
                _textpos
                for _bytepos, _textpos in byte_to_text_positions
                if target_bytepos <= _bytepos
            )
        # no newline to avoid scrolling on line 24
        console.list_line(bytes(output), newline=False, set_text_position=textpos)

    def renum(self, console, new_line, start_line, step):
        """Renumber stored program."""
        new_line = 10 if new_line is None else new_line
        start_line = 0 if start_line is None else start_line
        step = 10 if step is None else step
        # ensure we're not about to overwrite anything
        remaining = [_k for _k in self.line_numbers.keys() if _k < start_line]
        if remaining and new_line <= max(remaining):
            raise error.BASICError(error.IFC)
        # get a sorted list of line numbers
        # assign the new numbers
        old_to_new = {}
        for old_line in sorted(_k for _k in self.line_numbers.keys() if _k >= start_line):
            if old_line < 65535 and new_line > 65529:
                raise error.BASICError(error.IFC)
            if old_line == 65536:
                break
            old_to_new[old_line] = new_line
            self.last_stored = new_line
            new_line += step
        # write the new numbers
        for old_line in old_to_new:
            self.bytecode.seek(self.line_numbers[old_line])
            # skip the \x00\xC0\xDE & overwrite line number
            self.bytecode.read(3)
            self.bytecode.write(struct.pack('<H', old_to_new[old_line]))
        # write the indirect line numbers
        ins = self.bytecode
        ins.seek(0)
        while ins.skip_to_read((tk.T_UINT,)) == tk.T_UINT:
            # get the old g number
            token = ins.read(2)
            jumpnum, = struct.unpack('<H', token)
            # handle exception for ERROR GOTO
            if jumpnum == 0:
                pos = ins.tell()
                # skip line number token
                ins.seek(-3, 1)
                if ins.backskip_blank() == tk.GOTO and ins.backskip_blank() == tk.ERROR:
                    ins.seek(pos)
                    continue
                ins.seek(pos)
            try:
                newjump = old_to_new[jumpnum]
            except KeyError:
                # not redefined, exists in program?
                if jumpnum not in self.line_numbers:
                    linum = self.get_line_number(ins.tell()-1)
                    console.write_line(b'Undefined line %d in %d' % (jumpnum, linum))
                newjump = jumpnum
            ins.seek(-2, 1)
            ins.write(struct.pack('<H', newjump))
        # rebuild the line number dictionary
        new_lines = {}
        for old_line in old_to_new:
            new_lines[old_to_new[old_line]] = self.line_numbers[old_line]
            del self.line_numbers[old_line]
        self.line_numbers.update(new_lines)
        return old_to_new

    def load(self, g):
        """Load program from ascii, bytecode or protected stream."""
        self.erase()
        if g.filetype == b'B':
            # bytecode file
            self.bytecode.seek(1)
            self.bytecode.write(g.read())
        elif g.filetype == b'P':
            # protected file
            self.bytecode.seek(1)
            self.protected = self.allow_protect
            converter.unprotect(g, self.bytecode)
        elif g.filetype == b'A':
            # assume ASCII file
            # erase() only writes the terminator
            # so we need to get rid of the old code which is still in memory
            # or it'll end up after the new code in memory
            self.bytecode.truncate()
            # anything but numbers or whitespace: Direct Statement in File
            self.merge(g)
        else:
            logging.debug('Incorrect file type `%s` on LOAD', g.filetype)
        # rebuild line number dict and offsets
        if g.filetype != b'A':
            self.rebuild_line_dict()
        self.code_size = self.bytecode.tell()

    def merge(self, g):
        """Merge program from ascii or utf8 (if utf8_files is True) stream."""
        while True:
            line, cr = g.read_line()
            if not line and not cr:
                # end of file
                break
            elif cr is None:
                # line > 255 chars
                raise error.BASICError(error.LINE_BUFFER_OVERFLOW)
            linebuf = self.tokeniser.tokenise_line(line)
            if linebuf.read(1) == b'\0':
                # line starts with a number, add to program memory; store_line seeks to 1 first
                self.store_line(linebuf)
            else:
                # we have read the :
                if linebuf.skip_blank() not in tk.END_LINE:
                    raise error.BASICError(error.DIRECT_STATEMENT_IN_FILE)

    def save(self, g):
        """Save the program to stream g in (A)scii, (B)ytecode or (P)rotected mode."""
        mode = g.filetype
        if self.protected and mode != b'P':
            raise error.BASICError(error.IFC)
        current = self.bytecode.tell()
        # skip first \x00 in bytecode
        self.bytecode.seek(1)
        if mode == b'B':
            # binary bytecode mode
            g.write(self.bytecode.read())
        elif mode == b'P':
            # protected mode
            converter.protect(self.bytecode, g)
        else:
            # ascii mode
            while True:
                current_line, output, _ = self.lister.detokenise_line(self.bytecode)
                if current_line == -1 or (current_line > self.max_list_line):
                    break
                g.write_line(bytes(output))
        self.bytecode.seek(current)

    def list_lines(self, from_line, to_line):
        """List line range."""
        from_line, to_line = self.explicit_lines(from_line, to_line)
        if self.protected:
            # don't list protected files
            raise error.BASICError(error.IFC)
        # 65529 is max insertable line number for GW-BASIC 3.23.
        # however, 65530-65535 are executed if present in tokenised form.
        # in GW-BASIC, 65530 appears in LIST, 65531 and above are hidden
        if to_line is None:
            to_line = self.max_list_line
        numbers = [
            num for num in self.line_numbers
            if (from_line is None or num >= from_line) and num <= to_line
        ]
        # sort by positions, not line numbers!
        listable = sorted([self.line_numbers[num] for num in numbers])
        if numbers:
            self.last_stored = max(numbers)
        lines = []
        for pos in listable:
            self.bytecode.seek(pos + 1)
            _, line, _ = self.lister.detokenise_line(self.bytecode)
            lines.append(bytes(line))
        return lines

    def get_memory(self, offset):
        """Retrieve data from program code."""
        offset -= self.code_start
        code = self.bytecode.getvalue()
        try:
            return ord(code[offset:offset+1])
        except IndexError: # pragma: no cover
            # variable memory starts immediately after the stored program
            # ao memory access beyond the size of the program should not arrive here
            logging.debug('PEEK beyond program size handled by Program class.')
            return -1

    def get_memory_block(self, offset, length):
        """Retrieve block of data from program code."""
        offset -= self.code_start
        code = self.bytecode.getvalue()
        return bytearray(code[offset:offset+length])

    def set_memory(self, offset, val):
        """Change program code."""
        if not self.allow_code_poke:
            logging.warning('Ignored POKE into program code')
            return
        offset -= self.code_start
        loc = self.bytecode.tell()
        # move pointer to end
        self.bytecode.seek(0, 2)
        if offset > self.bytecode.tell(): # pragma: no cover
            # variable memory starts immediately after the stored program
            # ao memory access beyond the size of the program should not arrive here
            logging.debug('POKE beyond program size handled by Program class.')
        else:
            self.bytecode.seek(offset)
            self.bytecode.write(int2byte(val))
            # move pointer to end
            self.bytecode.seek(0, 2)
            self.rebuild_line_dict()
        # restore program pointer
        self.bytecode.seek(loc)



"""
PC-BASIC - display.modes
Emulated video modes

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from ..base import error
from ..base import bytematrix
from .colours import CGA2ColourMapper, CGA4ColourMapper, CGA16ColourMapper
from .colours import EGA16ColourMapper, EGA64ColourMapper
from .colours import EGA4ColourMapper, Tandy4ColourMapper, Olivetti2ColourMapper
from .colours import EGA16TextColourMapper, EGA64TextColourMapper
from .colours import MonoTextColourMapper, EGAMonoColourMapper, HerculesColourMapper
from .framebuffer import TextMemoryMapper, GraphicsMemoryMapper
from .framebuffer import CGAMemoryMapper, EGAMemoryMapper, Tandy6MemoryMapper
from .framebuffer import PackedTileBuilder, PlanedTileBuilder
from .framebuffer import PackedSpriteBuilder, PlanedSpriteBuilder, Tandy6SpriteBuilder


##############################################################################
# SCREEN modes by number for each adapter

_MODES = {
    'cga': {
        (0, 40): 'cgatext40',
        (0, 80): 'cgatext80',
        1: '320x200x4',
        2: '640x200x2',
    },
    'ega': {
        (0, 40): 'egatext40',
        (0, 80): 'egatext80',
        1: '320x200x4',
        2: '640x200x2',
        7: '320x200x16',
        8: '640x200x16',
        9: '640x350x16',
    },
    'ega_64k': {
        (0, 40): 'egatext40',
        (0, 80): 'egatext80',
        1: '320x200x4',
        2: '640x200x2',
        7: '320x200x16',
        8: '640x200x16',
        # 64k EGA adapter only has a 4-attr @ 16-colour palette in SCREEN 9
        9: '640x350x4c',
    },
    'vga': {
        (0, 40): 'vgatext40',
        (0, 80): 'vgatext80',
        1: '320x200x4',
        2: '640x200x2',
        7: '320x200x16',
        8: '640x200x16',
        9: '640x350x16',
    },
    'mda': {
        (0, 40): 'mdatext40',
        (0, 80): 'mdatext80',
    },
    'ega_mono': {
        (0, 40): 'ega_monotext40',
        (0, 80): 'ega_monotext80',
        10: '640x350x4',
    },
    'hercules': {
        (0, 40): 'mdatext40',
        (0, 80): 'mdatext80',
        3: '720x348x2',
    },
    'tandy': {
        (0, 40): 'tandytext40',
        (0, 80): 'tandytext80',
        1: '320x200x4_8pg',
        2: '640x200x2_8pg',
        3: '160x200x16',
        4: '320x200x4pcjr',
        5: '320x200x16pcjr',
        6: '640x200x4',
    },
    'pcjr': {
        (0, 40): 'cgatext40',
        (0, 80): 'cgatext80',
        1: '320x200x4_8pg',
        2: '640x200x2_8pg',
        3: '160x200x16',
        4: '320x200x4pcjr',
        5: '320x200x16pcjr',
        6: '640x200x4',
    },
    'olivetti': {
        (0, 40): 'olivettitext40',
        (0, 80): 'olivettitext80',
        1: '320x200x4',
        2: '640x200x2',
        3: '640x400x2',
    },
}
# on Olivetti M24, all numbers 3-255 give the same 'super resolution'/'altissima risoluzione'
_MODES['olivetti'].update({_mode: '640x400x2' for _mode in range(4, 256)})


##############################################################################
# screen number switches on WIDTH
# if we're in mode x, what mode y does WIDTH w take us to? {x: {w: y}}
TO_WIDTH = {
    'cga': {
        0: {40: 0, 80: 0},
        1: {40: 1, 80: 2},
        2: {40: 1, 80: 2},
    },
    'ega': {
        0: {40: 0, 80: 0},
        1: {40: 1, 80: 2},
        2: {40: 1, 80: 2},
        7: {40: 7, 80: 8},
        8: {40: 7, 80: 8},
        9: {40: 1, 80: 9},
    },
    'mda': {
        0: {40: 0, 80: 0},
    },
    'ega_mono': {
        0: {40: 0, 80: 0},
        10: {40: 0, 80: 10},
    },
    'hercules': {
        0: {40: 0, 80: 0},
        3: {40: 0, 80: 3},
    },
    'pcjr': {
        0: {20: 3, 40: 0, 80: 0},
        1: {20: 3, 40: 1, 80: 2},
        2: {20: 3, 40: 1, 80: 2},
        3: {20: 3, 40: 1, 80: 2},
        4: {20: 3, 40: 4, 80: 2},
        5: {20: 3, 40: 5, 80: 6},
        6: {20: 3, 40: 5, 80: 6},
    },
    'olivetti': {
        0: {40: 0, 80: 0},
        1: {40: 1, 80: 2},
        2: {40: 1, 80: 2},
        3: {40: 1, 80: 3}, # assumption
    },
}

# also look up graphics modes by name
for _adapter in TO_WIDTH:
    TO_WIDTH[_adapter].update({
        _MODES[_adapter][_nr]: _switches for _nr, _switches in TO_WIDTH[_adapter].items() if _nr
    })

TO_WIDTH['ega_64k'] = TO_WIDTH['ega']
TO_WIDTH['vga'] = TO_WIDTH['ega']
TO_WIDTH['tandy'] = TO_WIDTH['pcjr']


def to_width(adapter, mode, to_width):
    """Get target mode number for WIDTH change."""
    if mode.is_text_mode and to_width in (40, 80):
        return 0
    else:
        try:
            return TO_WIDTH[adapter][mode.name][to_width]
        except KeyError:
            # raise an error if the width value doesn't make sense
            raise error.BASICError(error.IFC)


##############################################################################
# video mode number

_MODE_NUMBER = {
    '640x200x2': 6, '160x200x16': 8, '320x200x16pcjr': 9,
    '640x200x4': 10, '320x200x16': 13, '640x200x16': 14,
    '640x350x4': 15, '640x350x16': 16, '640x350x4c': 16, '640x400x2': 0x40,
    '320x200x4pcjr': 4, '320x200x4': 4
    # '720x348x2': ? # hercules - unknown
}

def get_mode_number(mode, colorswitch):
    """Get the low-level mode number used by mode switching interrupt."""
    if mode.is_text_mode:
        if mode.name in ('mdatext80', 'ega_monotext80'):
            return 7
        return (mode.width == 40) * 2 + colorswitch % 2
    elif mode.name == '320x200x4':
        return 4 + colorswitch % 2 # mode_5 if colorswitch == 1
    else:
        try:
            return _MODE_NUMBER[mode.name]
        except KeyError:
            return 0xff


##############################################################################
# video mode factory

def get_mode(number, width, adapter, monitor, video_mem_size):
    """Retrieve text or graphical mode by screen number."""
    try:
        if number:
            name = _MODES[adapter][number]
        else:
            name = _MODES[adapter][0, width]
    except KeyError:
        # no such mode
        raise error.BASICError(error.IFC)
    mode_data = dict(**_MODE_INFO[name])
    cls = mode_data.pop('layout')
    return cls(name=name, video_mem_size=video_mem_size, **mode_data)


##############################################################################
# video mode base class

class VideoMode(object):
    """Base class for video modes."""

    is_text_mode = None
    is_cga_hires = False

    def __init__(
            self, name, height, width, font_height, font_width, attr, cursor_shape, colourmap
        ):
        """Initialise video mode settings."""
        self.name = name
        self.height = height
        self.width = width
        self.font_height = font_height
        self.font_width = font_width
        self.pixel_height = height * font_height
        self.pixel_width = width * font_width
        # initial/default attribute
        self.attr = attr
        # override these
        self.memorymap = None
        self.cursor_shape = cursor_shape
        self.colourmap = colourmap

    @property
    def num_pages(self):
        """Number of available pages."""
        return self.memorymap.num_pages

    def __eq__(self, rhs):
        """Equality with another mode or mode name."""
        if isinstance(rhs, VideoMode):
            return self.name == rhs.name
        else:
            return self.name == rhs

    def __ne__(self, rhs):
        """Inequality with another mode or mode name."""
        return not self.__eq__(rhs)


##############################################################################
# text modes

class TextMode(VideoMode):
    """Default settings for a text mode."""

    _textmemorymapper = TextMemoryMapper
    is_text_mode = True

    def __init__(
            self, name, rows, columns, font_height, font_width, attr, cursor_shape,
            video_mem_size, max_pages, mono, colourmap
        ):
        """Initialise video mode settings."""
        VideoMode.__init__(
            self, name, rows, columns, font_height, font_width, attr, cursor_shape, colourmap
        )
        self.memorymap = self._textmemorymapper(rows, columns, video_mem_size, max_pages, mono)


##############################################################################
# graphics modes


class GraphicsMode(VideoMode):
    """Default settings for a graphics mode."""

    # override these
    _memorymapper = GraphicsMemoryMapper
    _tile_builder = lambda _: None
    _sprite_builder = lambda _: None

    # text mode flag
    is_text_mode = False

    def __init__(
            self, name, width, height, rows, columns,
            attr, bitsperpixel, interleave_times, bank_size, video_mem_size, max_pages,
            cursor_attr, colourmap, is_cga_hires=False
        ):
        """Initialise video mode settings."""
        font_width = width // columns
        # ceildiv for hercules (14-px font on 348 lines)
        font_height = -(-height // rows)
        # block cursor in graphics mode
        cursor_shape = (0, font_height-1)
        VideoMode.__init__(
            self, name, rows, columns, font_height, font_width, attr, cursor_shape, colourmap
        )
        self.is_cga_hires = is_cga_hires
        # override pixel dimensions
        self.pixel_height = height
        self.pixel_width = width
        # used in display.py to initialise pixelbuffer
        self.bitsperpixel = bitsperpixel
        self.memorymap = self._memorymapper(
            height, width, video_mem_size, max_pages, interleave_times, bank_size, bitsperpixel
        )
        # cursor attribute
        self.cursor_attr = cursor_attr
        # sprite and tile builders
        self.build_tile = self._tile_builder(self.bitsperpixel)
        self.sprite_builder = self._sprite_builder(self.bitsperpixel)


class CGAMode(GraphicsMode):
    """Default settings for a CGA graphics mode."""

    _memorymapper = CGAMemoryMapper
    _tile_builder = PackedTileBuilder
    _sprite_builder = PackedSpriteBuilder


class EGAMode(GraphicsMode):
    """Default settings for a EGA graphics mode."""

    _memorymapper = EGAMemoryMapper
    _tile_builder = PlanedTileBuilder
    _sprite_builder = PlanedSpriteBuilder

    def __init__(
            self, name, width, height, rows, columns,
            attr, bitsperpixel, interleave_times, bank_size, video_mem_size, max_pages,
            cursor_attr, colourmap, planes_used=range(4)
        ):
        """Initialise video mode settings."""
        GraphicsMode.__init__(
            self, name, width, height, rows, columns,
            attr, bitsperpixel, interleave_times, bank_size, video_mem_size, max_pages,
            cursor_attr, colourmap
        )
        # EGA memorymap settings
        self.memorymap.set_planes_used(planes_used)


class Tandy6Mode(GraphicsMode):
    """Default settings for Tandy graphics mode 6."""

    _memorymapper = Tandy6MemoryMapper
    _tile_builder = PackedTileBuilder
    # initialising this with self.bitsperpixel should do the right thing
    _sprite_builder = Tandy6SpriteBuilder



##############################################################################
# mode definitions

_MODE_INFO = {
    # text modes

    'vgatext40': dict(
        rows=25, columns=40, font_height=16, font_width=9, attr=7, max_pages=8, mono=False,
        cursor_shape=(15, 15), layout=TextMode, colourmap=EGA64TextColourMapper
    ),
    'vgatext80': dict(
        rows=25, columns=80, font_height=16, font_width=9, attr=7, max_pages=4, mono=False,
        cursor_shape=(15, 15), layout=TextMode, colourmap=EGA64TextColourMapper
    ),
    'egatext40': dict(
        rows=25, columns=40, font_height=14, font_width=8, attr=7, max_pages=8, mono=False,
        cursor_shape=(12, 12), layout=TextMode, colourmap=EGA64TextColourMapper
    ),
    'egatext80': dict(
        rows=25, columns=80, font_height=14, font_width=8, attr=7, max_pages=4, mono=False,
        cursor_shape=(12, 12), layout=TextMode, colourmap=EGA64TextColourMapper
    ),
    'ega_monotext40': dict(
        rows=25, columns=40, font_height=14, font_width=8, attr=7, max_pages=8, mono=True,
        cursor_shape=(12, 12), layout=TextMode, colourmap=MonoTextColourMapper
    ),
    'ega_monotext80': dict(
        rows=25, columns=80, font_height=14, font_width=8, attr=7, max_pages=4, mono=True,
        cursor_shape=(12, 12), layout=TextMode, colourmap=MonoTextColourMapper
    ),
    'mdatext40': dict(
        rows=25, columns=40, font_height=14, font_width=9, attr=7, max_pages=1, mono=True,
        cursor_shape=(13, 13), layout=TextMode, colourmap=MonoTextColourMapper
    ),
    'mdatext80': dict(
        rows=25, columns=80, font_height=14, font_width=9, attr=7, max_pages=1, mono=True,
        cursor_shape=(13, 13), layout=TextMode, colourmap=MonoTextColourMapper
    ),
    'tandytext40': dict(
        rows=25, columns=40, font_height=9, font_width=8, attr=7, max_pages=8, mono=False,
        cursor_shape=(7, 7), layout=TextMode, colourmap=EGA16TextColourMapper
    ),
    'tandytext80': dict(
        rows=25, columns=80, font_height=9, font_width=8, attr=7, max_pages=4, mono=False,
        cursor_shape=(7, 7), layout=TextMode, colourmap=EGA16TextColourMapper
    ),
    'cgatext40': dict(
        rows=25, columns=40, font_height=8, font_width=8, attr=7, max_pages=8, mono=False,
        cursor_shape=(7, 7), layout=TextMode, colourmap=EGA16TextColourMapper
    ),
    'cgatext80': dict(
        rows=25, columns=80, font_height=8, font_width=8, attr=7, max_pages=4, mono=False,
        cursor_shape=(7, 7), layout=TextMode, colourmap=EGA16TextColourMapper
    ),
    'olivettitext40': dict(
        rows=25, columns=40, font_height=16, font_width=8, attr=7, max_pages=8, mono=False,
        cursor_shape=(15, 15), layout=TextMode, colourmap=EGA16TextColourMapper
    ),
    'olivettitext80': dict(
        rows=25, columns=80, font_height=16, font_width=8, attr=7, max_pages=4, mono=False,
        cursor_shape=(15, 15), layout=TextMode, colourmap=EGA16TextColourMapper
    ),

    # graphics modes

    '320x200x4': dict(
        # 04h 320x200x4  16384B 2bpp 0xb8000    screen 1
        # cga/ega: 1 page only
        width=320, height=200, rows=25, columns=40, attr=3, cursor_attr=None,
        bitsperpixel=2, interleave_times=2, bank_size=0x2000, max_pages=1,
        layout=CGAMode, colourmap=CGA4ColourMapper
    ),
    '640x200x2': dict(
        # 06h 640x200x2  16384B 1bpp 0xb8000    screen 2
        width=640, height=200, rows=25, columns=80, attr=1, cursor_attr=None,
        bitsperpixel=1, interleave_times=2, bank_size=0x2000, max_pages=1,
        layout=CGAMode, colourmap=CGA2ColourMapper, is_cga_hires=True
    ),
    '160x200x16': dict(
        # 08h 160x200x16 16384B 4bpp 0xb8000    PCjr/Tandy screen 3
        width=160, height=200, rows=25, columns=20, attr=15, cursor_attr=3,
        bitsperpixel=4, interleave_times=2, bank_size=0x2000, max_pages=8,
        layout=CGAMode, colourmap=CGA16ColourMapper
    ),
    '320x200x4pcjr': dict(
        #     320x200x4  16384B 2bpp 0xb8000   Tandy/PCjr screen 4
        width=320, height=200, rows=25, columns=40, attr=3, cursor_attr=3,
        bitsperpixel=2, interleave_times=2, bank_size=0x2000, max_pages=8,
        layout=CGAMode, colourmap=Tandy4ColourMapper
    ),
    '320x200x16pcjr': dict(
        # 09h 320x200x16 32768B 4bpp 0xb8000    Tandy/PCjr screen 5
        width=320, height=200, rows=25, columns=40, attr=15, cursor_attr=3,
        bitsperpixel=4, interleave_times=4, bank_size=0x2000, max_pages=4,
        layout=CGAMode, colourmap=CGA16ColourMapper
    ),
    '640x200x4': dict(
        # 0Ah 640x200x4  32768B 2bpp 0xb8000   Tandy/PCjr screen 6
        width=640, height=200, rows=25, columns=80, attr=3, cursor_attr=3,
        bitsperpixel=2, interleave_times=4, bank_size=0x2000, max_pages=4,
        layout=Tandy6Mode, colourmap=Tandy4ColourMapper
    ),
    '320x200x16': dict(
        # 0Dh 320x200x16 32768B 4bpp 0xa0000    EGA screen 7
        width=320, height=200, rows=25, columns=40, attr=15, cursor_attr=None,
        bitsperpixel=4, interleave_times=1, bank_size=0x2000, max_pages=None,
        layout=EGAMode, colourmap=EGA16ColourMapper
    ),
    '640x200x16': dict(
        # 0Eh 640x200x16    EGA screen 8
        width=640, height=200, rows=25, columns=80, attr=15, cursor_attr=None,
        bitsperpixel=4, interleave_times=1, bank_size=0x4000, max_pages=None,
        layout=EGAMode, colourmap=EGA16ColourMapper
    ),
    '640x350x16': dict(
        # 10h 640x350x16    EGA screen 9
        width=640, height=350, rows=25, columns=80, attr=15, cursor_attr=None,
        bitsperpixel=4, interleave_times=1, bank_size=0x8000, max_pages=None,
        layout=EGAMode, colourmap=EGA64ColourMapper
    ),
    '640x350x4c': dict(
        # 10h 640x350x16    EGA screen 9 with 64K
        width=640, height=350, rows=25, columns=80, attr=3, cursor_attr=None,
        bitsperpixel=2, interleave_times=1, bank_size=0x8000, max_pages=None,
        layout=EGAMode, colourmap=EGA4ColourMapper
    ),
    '640x350x4': dict(
        # 0Fh 640x350x4     EGA monochrome screen 10
        width=640, height=350, rows=25, columns=80, attr=1, cursor_attr=None,
        bitsperpixel=2, interleave_times=1, bank_size=0x8000, max_pages=None, planes_used=(1, 3),
        layout=EGAMode, colourmap=EGAMonoColourMapper
    ),
    '640x400x2': dict(
        # 40h 640x400x2   1bpp  olivetti screen 3-255
        width=640, height=400, rows=25, columns=80, attr=1, cursor_attr=None,
        bitsperpixel=1, interleave_times=4, bank_size=0x2000, max_pages=1,
        layout=CGAMode, colourmap=Olivetti2ColourMapper
    ),
    '720x348x2': dict(
        # hercules screen 3
        # SCREEN 3 supports two pages (0 and 1);
        # SCREEN 0 used with Hercules supports only one page.
        # see MS KB 21839, https://jeffpar.github.io/kbarchive/kb/021/Q21839/
        width=720, height=348, rows=25, columns=80, attr=1, cursor_attr=None,
        bitsperpixel=1, interleave_times=4, bank_size=0x2000, max_pages=2,
        layout=CGAMode, colourmap=HerculesColourMapper
    ),
}

# tandy/pcjr 8-page versions of standard CGA modes (?)
_MODE_INFO['320x200x4_8pg'] = _MODE_INFO['320x200x4']
_MODE_INFO['640x200x2_8pg'] = _MODE_INFO['640x200x2']
_MODE_INFO['320x200x4_8pg']['max_pages'] = 8
_MODE_INFO['640x200x2_8pg']['max_pages'] = 8



"""
PC-BASIC - sound.py
Sound handling

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from collections import deque
import datetime

from ..compat import iterchar, zip
from .base import error
from .base import signals
from .base import tokens as tk
from .base.tokens import DIGITS
from . import mlparser
from . import values


# NOTE - sound timings and queue lengths are fairly close to DOSBox for Tandy.
# For regular GW-BASIC there are some differences
# and the results of the original are difficult to understand


# number of tones, gaps or markers in background buffer
BACKGROUND_BUFFER_LENGTH = 32

# base frequency for noise source
BASE_FREQ = 3579545. / 1024.
# frequency multipliers for noise sources 0-7
NOISE_FREQ = tuple(BASE_FREQ * v for v in (1., 0.5, 0.25, 0., 1., 0.5, 0.25, 0.))

# 12-tone equal temperament
# C, C#, D, D#, E, F, F#, G, G#, A, A#,
NOTE_FREQ = tuple(440. * 2**((i-33.)/12.) for i in range(84))
NOTES = {
    b'C': 0, b'C#': 1, b'D-': 1, b'D': 2, b'D#': 3, b'E-': 3, b'E': 4, b'F': 5, b'F#': 6,
    b'G-': 6, b'G': 7, b'G#': 8, b'A-': 8, b'A': 9, b'A#': 10, b'B-': 10, b'B': 11
}

# critical duration value below which sound loops
# in BASIC, 1/44 = 0.02272727248 which is '\x8c\x2e\x3a\x7b'
LOOP_THRESHOLD = 0.02272727248

# length of a clock tick ("PIT tick", see Joel Yliluoma's noise.bas)
TICK_LENGTH = 0x1234DC / 65536.


class Sound(object):
    """Sound queue manipulations."""

    def __init__(self, queues, values, memory, syntax):
        """Initialise sound queue."""
        # for wait() and queues
        self._queues = queues
        self._values = values
        self._memory = memory
        # Tandy/PCjr noise generator
        # frequency for noise sources
        self._noise_freq = list(NOISE_FREQ)
        # advnced sound capabilities
        self._multivoice = syntax if syntax in ('pcjr', 'tandy') else b''
        # Tandy/PCjr SOUND ON and BEEP ON
        # tandy has SOUND ON by default, pcjr has it OFF
        self._sound_on = (self._multivoice == 'tandy')
        # pc-speaker on/off; (not implemented; not sure whether should be on)
        self._beep_on = True
        # timed queues for each voice (including gaps, for background counting & rebuilding)
        self._voice_queue = [TimedQueue(), TimedQueue(), TimedQueue(), TimedQueue()]
        self._foreground = True
        self._synch = False
        # initialise PLAY state
        self.reset_play()

    @property
    def multivoice(self):
        """We have multivoice capability."""
        return bool(self._multivoice)

    def beep_(self, args):
        """BEEP: produce an alert sound or switch internal speaker on/off."""
        command, = args
        if command:
            self._beep_on = (command == tk.ON)
        else:
            self.beep()

    def beep(self):
        """Produce an alert sound."""
        self.emit_tone(800, 0.25, fill=1, loop=False, voice=0, volume=15)
        # at most 16 notes in the sound queue with gaps, or 32 without gaps
        self._wait_background()

    def emit_tone(self, frequency, duration, fill, loop, voice, volume):
        """Play a sound on the tone generator."""
        if frequency < 0:
            frequency = 0
        if self._multivoice and self._sound_on and 0 < frequency < 110.:
            # pcjr, tandy play low frequencies as 110Hz
            frequency = 110.
        # if a synch is requested, emit it at the first tone or pause
        if self._synch:
            self.emit_synch()
        # no sound if both SOUND OFF and BEEP OFF - note that dosbox gets this wrong
        # checked on real hardware, see https://www.vogons.org/viewtopic.php?t=56735&p=626006
        if not (self._beep_on or self._sound_on):
            volume = 0
        tone = signals.Event(signals.AUDIO_TONE, (voice, frequency, fill*duration, loop, volume))
        self._queues.audio.put(tone)
        self._voice_queue[voice].put(tone, None if loop else fill*duration, True)
        # separate gap event, except for legato (fill==1)
        if fill != 1 and not loop:
            gap = signals.Event(signals.AUDIO_TONE, (voice, 0, (1-fill) * duration, 0, 0))
            self._queues.audio.put(gap)
            self._voice_queue[voice].put(gap, (1-fill) * duration, False)
        if voice == 2 and frequency != 0:
            # reset linked noise frequencies
            # /2 because we're using a 0x4000 rotation rather than 0x8000
            self._noise_freq[3] = frequency / 2.
            self._noise_freq[7] = frequency / 2.

    def emit_noise(self, source, volume, duration, loop):
        """Generate a noise."""
        frequency = self._noise_freq[source]
        # if not SOUND ON an IFC was raised, so don't check here
        noise = signals.Event(signals.AUDIO_NOISE, (source > 3, frequency, duration, loop, volume))
        self._queues.audio.put(noise)
        self._voice_queue[3].put(noise, None if loop else duration, True)

    def sound_(self, args):
        """SOUND: produce a sound or switch external speaker on/off."""
        arg0 = next(args)
        if self._multivoice and arg0 in (tk.ON, tk.OFF):
            command = arg0
        else:
            command = None
            freq = values.to_int(arg0)
            dur = values.to_single(next(args)).to_value()
            error.range_check(-65535, 65535, dur)
            volume = next(args)
            if volume is None:
                volume = 15
            else:
                volume = values.to_int(volume)
                error.range_check(-1, 15, volume)
                if volume == -1:
                    volume = 15
            voice = next(args)
            if voice is None:
                voice = 0
            else:
                # on PCjr, four-parameter SOUND requires SOUND ON
                if self._multivoice == 'pcjr' and not self._sound_on:
                    raise error.BASICError(error.IFC)
                voice = values.to_int(voice)
                error.range_check(0, 2, voice) # can't address noise channel here
        list(args)
        if command is not None:
            self._sound_on = (command == tk.ON)
            self.stop_all_sound()
            return
        if dur == 0:
            self.stop_all_sound()
            return
        # Tandy only allows frequencies below 37 (but plays them as 110 Hz)
        if freq != 0:
            # 32767 is pause
            error.range_check(-32768 if self._multivoice == 'tandy' else 37, 32767, freq)
        # calculate duration in seconds
        dur_sec = dur / TICK_LENGTH
        # loop if duration less than 1/44 == 0.02272727248
        if dur < LOOP_THRESHOLD:
            # play indefinitely in background
            self.emit_tone(freq, dur_sec, fill=1, loop=True, voice=voice, volume=volume)
            self._wait_background()
        else:
            self.emit_tone(freq, dur_sec, fill=1, loop=False, voice=voice, volume=volume)
            if self._foreground:
                # continue when last tone has started playing, both on tandy and gw
                # this is different from what PLAY does!
                self._wait(1)
            else:
                self._wait_background()

    def noise_(self, args):
        """Generate a noise (NOISE statement)."""
        if not self._sound_on:
            raise error.BASICError(error.IFC)
        source = values.to_int(next(args))
        error.range_check(0, 7, source)
        volume = values.to_int(next(args))
        error.range_check(0, 15, volume)
        dur = values.to_single(next(args)).to_value()
        error.range_check(-65535, 65535, dur)
        list(args)
        # calculate duration in seconds
        dur_sec = dur / TICK_LENGTH
        # loop if duration less than 1/44 == 0.02272727248
        self.emit_noise(source, volume, dur_sec, loop=(dur < LOOP_THRESHOLD))
        # noise is always background
        self._wait_background()

    def _wait_background(self):
        """Wait until the background queue becomes available."""
        # 32 plus one playing
        self._wait(BACKGROUND_BUFFER_LENGTH+1)

    def _wait(self, wait_length):
        """Wait until queue is shorter than or equal to given length."""
        # top of queue is the currently playing tone or gap
        while max(len(queue) for queue in self._voice_queue) > wait_length:
            self._queues.wait()

    def stop_all_sound(self):
        """Terminate all sounds immediately."""
        for q in self._voice_queue:
            q.clear()
        self._queues.audio.put(signals.Event(signals.AUDIO_STOP))

    def persist(self, flag):
        """Set mixer persistence flag (runmode)."""
        self._queues.audio.put(signals.Event(signals.AUDIO_PERSIST, (flag,)))

    def rebuild(self):
        """Rebuild tone queues."""
        # should we pop one at a time from each voice queue to equalise timings?
        for voice, q in enumerate(self._voice_queue):
            for item, duration in q.items():
                item.params = list(item.params)
                item.params[2] = duration
                self._queues.audio.put(item)

    def play_fn_(self, args):
        """PLAY function: get length of music queue."""
        voice = values.to_int(next(args))
        list(args)
        error.range_check(0, 255, voice)
        if not(self._multivoice and voice in (1, 2)):
            voice = 0
        return self._values.new_integer().from_int(self._voice_queue[voice].tones_waiting())

    def tones_waiting(self):
        """Return max number of tones waiting in queues."""
        return max(self._voice_queue[voice].tones_waiting() for voice in range(3))

    def emit_synch(self):
        """Synchronise the three tone voices."""
        # on Tandy/PCjr, align voices (excluding noise) at the end of each PLAY statement
        if self._multivoice:
            max_time = max(q.expiry() for q in self._voice_queue[:3])
            for voice, q in enumerate(self._voice_queue[:3]):
                duration = (max_time - q.expiry()).total_seconds()
                # fill up the queue with the necessary amount of silence
                # this takes up one spot in the buffer and thus affects timings
                # which is intentional
                balloon = signals.Event(signals.AUDIO_TONE, (voice, 0, duration, False, 0))
                self._queues.audio.put(balloon)
                self._voice_queue[voice].put(balloon, duration, None)
        self._synch = False

    def reset_play(self):
        """Reset PLAY state."""
        # music foreground (MF) mode
        self._foreground = True
        # reset all PLAY state
        self._state = [PlayState(), PlayState(), PlayState()]

    def play_(self,  args):
        """Parse a list of Music Macro Language strings (PLAY statement)."""
        # retrieve Music Macro Language string
        mml_list = [values.to_string_or_none(arg) for arg, _ in zip(args, range(3))]
        list(args)
        # at least one string must be specified
        if not any(mml_list):
            raise error.BASICError(error.MISSING_OPERAND)
        # on PCjr, three-voice PLAY requires SOUND ON
        if (self._multivoice == 'pcjr') and not self._sound_on and len(mml_list) > 1:
            raise error.BASICError(error.STX)
        # a marker is inserted at the start of the PLAY statement
        # this takes up one spot in the buffer and thus affects timings
        self._synch = True
        mml_list += [b''] * (3 - len(mml_list))
        ml_parser_list = [mlparser.MLParser(mml, self._memory, self._values) for mml in mml_list]
        next_oct = 0
        voices = list(range(3))
        while True:
            if not voices:
                break
            for voice in voices:
                vstate = self._state[voice]
                mmls = ml_parser_list[voice]
                c = mmls.skip_blank_read().upper()
                if c == b'':
                    voices.remove(voice)
                    continue
                if c == b';':
                    # absorb one (and only one) semicolon
                    c = mmls.skip_blank_read().upper()
                if c == b'X':
                    # insert substring
                    sub = mmls.parse_string()
                    pos = mmls.tell()
                    rest = mmls.read()
                    mmls.seek(pos)
                    mmls.truncate()
                    mmls.write(sub)
                    mmls.write(rest)
                    mmls.seek(pos)
                elif c == b'N':
                    note = mmls.parse_number()
                    error.range_check(0, 84, note)
                    dur = vstate.length
                    while mmls.skip_blank_read_if((b'.',)):
                        dur *= 1.5
                    if note == 0:
                        # pause
                        self.emit_tone(0, dur*vstate.tempo, 1, False, voice, vstate.volume)
                    else:
                        self.emit_tone(
                                NOTE_FREQ[note-1], dur*vstate.tempo,
                                vstate.fill, False, voice, vstate.volume)
                elif c == b'L':
                    recip = mmls.parse_number()
                    error.range_check(1, 64, recip)
                    vstate.length = 1. / recip
                elif c == b'T':
                    recip = mmls.parse_number()
                    error.range_check(32, 255, recip)
                    vstate.tempo = 240. / recip
                elif c == b'O':
                    octave = mmls.parse_number()
                    error.range_check(0, 6, octave)
                    vstate.octave = octave
                elif c == b'>':
                    vstate.octave += 1
                    if vstate.octave > 6:
                        vstate.octave = 6
                elif c == b'<':
                    vstate.octave -= 1
                    if vstate.octave < 0:
                        vstate.octave = 0
                elif c in (b'A', b'B', b'C', b'D', b'E', b'F', b'G', b'P'):
                    note = c
                    dur = vstate.length
                    length = None
                    if mmls.skip_blank_read_if((b'#', b'+')):
                        note += b'#'
                    elif mmls.skip_blank_read_if((b'-',)):
                        note += b'-'
                    c = mmls.skip_blank_read_if(DIGITS)
                    if c is not None:
                        numstr = [c]
                        while mmls.skip_blank() in set(iterchar(DIGITS)):
                            numstr.append(mmls.read(1))
                        # NOT ml_parse_number, only literals allowed here!
                        length = int(b''.join(numstr))
                        error.range_check(0, 64, length)
                        if length > 0:
                            dur = 1. / float(length)
                    while mmls.skip_blank_read_if((b'.',)):
                        error.throw_if(note == b'P' and length == 0)
                        dur *= 1.5
                    if note == b'P':
                        # length must be specified
                        if length is None:
                            raise error.BASICError(error.IFC)
                        # don't do anything for length 0
                        elif length > 0:
                            self.emit_tone(0, dur * vstate.tempo, 1, False, voice, vstate.volume)
                    else:
                        # use default length for length 0
                        try:
                            self.emit_tone(
                                NOTE_FREQ[(vstate.octave + next_oct) * 12 + NOTES[note]],
                                dur * vstate.tempo, vstate.fill, False, voice, vstate.volume)
                        except KeyError:
                            raise error.BASICError(error.IFC)
                    next_oct = 0
                elif c == b'M':
                    c = mmls.skip_blank_read().upper()
                    if c == b'N':
                        vstate.fill = 7./8.
                    elif c == b'L':
                        vstate.fill = 1.
                    elif c == b'S':
                        vstate.fill = 3./4.
                    elif c == b'F':
                        self._foreground = True
                    elif c == b'B':
                        self._foreground = False
                    else:
                        raise error.BASICError(error.IFC)
                elif c == b'V' and (
                        self._multivoice and self._sound_on
                        or self._multivoice == 'tandy'
                    ):
                    vol = mmls.parse_number()
                    error.range_check(-1, 15, vol)
                    if vol == -1:
                        vstate.volume = 15
                    else:
                        vstate.volume = vol
                else:
                    raise error.BASICError(error.IFC)
        self._synch = False
        if self._foreground:
            # wait until fully done on Tandy/PCjr, continue early on GW
            self._wait(0 if self._multivoice else 1)
        else:
            self._wait_background()


class PlayState(object):
    """State variables of the PLAY command."""

    def __init__(self):
        """Initialise play state."""
        self.octave = 4
        self.fill = 7./8.
        # 2*0.25 = 0.5 seconds per quarter note
        self.tempo = 2.
        self.length = 0.25
        self.volume = 15


###############################################################################
# sound queue

class TimedQueue(object):
    """Queue with expiring elements."""

    def __init__(self):
        """Initialise timed queue."""
        self._deque = deque()
        # hack to reproduce queue lengths as reported by GW-BASIC
        self._balloon_popped = False

    def __getstate__(self):
        """Get pickling dict for queue."""
        self._check_expired()
        return {
            'deque': self._deque,
            'now': datetime.datetime.now(),
            'balloon_popped': self._balloon_popped,
        }

    def __setstate__(self, st):
        """Initialise queue from pickling dict."""
        offset = datetime.datetime.now() - st['now']
        self._deque = deque((item, expiry+offset, counts) for (item, expiry, counts) in st['deque'])
        self._balloon_popped = st['balloon_popped']

    def _check_expired(self):
        """Drop expired items from queue."""
        counts = 0
        try:
            while self._deque[0][1] <= datetime.datetime.now():
                popped = self._deque.popleft()
                self._balloon_popped = (popped[2] is None)
        except (IndexError, TypeError):
            pass

    def put(self, item, duration, count_for_size):
        """
        Put item onto queue with duration in seconds.
        Items with duration None remain until next item is put.
        """
        self._check_expired()
        # drop looping elements
        try:
            if self._deque[-1][1] is None:
                self._deque.pop()
        except IndexError:
            pass
        if duration is None:
            expiry = None
        else:
            last = self._deque[-1][1] if self._deque else datetime.datetime.now()
            expiry = max(last, datetime.datetime.now()) + datetime.timedelta(seconds=duration)
        self._deque.append((item, expiry, count_for_size))

    def clear(self):
        """Clear the queue."""
        self._deque.clear()

    def __len__(self):
        """Number of elements in queue."""
        self._check_expired()
        return len(self._deque)

    def tones_waiting(self):
        """Number of tones (not gaps) waiting in queue."""
        self._check_expired()
        # count number of notes waiting, exclude the top of queue ("now playing")
        waiting = len([item for i, item in enumerate(self._deque) if item[2] and i])
        # hack: if the most recent item popped was a balloon
        # (i.e. we've just started a PLAY and the first note has not finished)
        # include the first note in the waiting queue length
        waiting += self._balloon_popped
        self._balloon_popped = False
        return waiting

    def expiry(self):
        """Last expiry in queue, return now() for looping sound."""
        self._check_expired()
        try:
            return self._deque[-1][1] or datetime.datetime.now()
        except IndexError:
            return datetime.datetime.now()

    def items(self):
        """Iterate over each item and its duration."""
        self._check_expired()
        last_expiry = datetime.datetime.now()
        for item, expiry, _ in self._deque:
            if expiry is None:
                duration = None
            else:
                # adjust duration
                duration = (expiry - last_expiry).total_seconds()
                last_expiry = expiry
            yield item, duration
